<'
# Copyright (c) 2012-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Thu Mar 15 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include "portLib.h"
#include "errnoLib.h"
#include "gcomLib.h"

#include "<"$comp">/c/client.h"
#include "serialize.h"
#include "typecopy.h"
#include "<"$comp">_msglib.h"

<'foreach s [$component services] {'>

/* <"[--- Service [$s name] encoding/decoding ---------------------------]"> */

int
genom_<"$comp">_<"[$s name]">_encode(
  char *buffer, int size, char *dst, int maxsize)
{
<'  if {[llength [$s parameters in inout]]} {'>
  char *p = dst;
  struct genom_<"$comp">_<"[$s name]">_input *in =
    (struct genom_<"$comp">_<"[$s name]">_input *)buffer;
  size_t s;

  s = 0;
<'    foreach p [$s parameters in inout] {'>
  s += genom_serialen_<"[[$p type] mangle]">(
    <"[[$p type] pass value in->[$p name]]">);
<'    }'>
  if (s > maxsize) return ERROR;

<'    foreach p [$s parameters in inout] {'>
  genom_serialize_<"[[$p type] mangle]">(
    &p, <"[[$p type] pass value in->[$p name]]">);
<'    }'>

  return s;
<'  } else {'>
  return 0;
<'  }'>
}

int
genom_<"$comp">_<"[$s name]">_decode(
  char *buffer, int size, char *dst, int maxsize)
{
<'  if {[llength [$s parameters inout out]]} {'>
  struct genom_<"$comp">_<"[$s name]">_output *out =
    (struct genom_<"$comp">_<"[$s name]">_output *)dst;
<'  }'>
  ssize_t len = size;
  char *p = buffer;
  int s;
  if (size == 0) return 0;

  if (*p == *"") {
    p++; len--;

<'  foreach p [$s parameters inout out] {'>
    s = genom_deserialize_<"[[$p type] mangle]">(
      &p, &len, <"[[$p type] pass reference out->[$p name]]">);
<'  }'>
<'    foreach e [$s throws] {'>
  } else if (!strcmp(p, <"[$e cname]">_id)) {
<'      if {[llength [$e members]]} {'>
    <"[$e declarator]"> exdetail;
    p += sizeof(<"[$e cname]">_id); len -= sizeof(<"[$e cname]">_id);
    genom_tinit_<"[$e mangle]">(&exdetail);
    s = genom_deserialize_<"[$e mangle]">(&p, &len, &exdetail);
    if (!s) <"[$e cname]">(&exdetail);
<'      } else {'>
    <"[$e cname]">();
    s = 0; len -= sizeof(<"[$e cname]">_id);
<'      }'>
    errnoSet(ERROR);
    size = ERROR;
<'    }'>
  } else if (!strcmp(p, genom_syserr_id)) {
    genom_syserr_detail exdetail;
    p += sizeof(genom_syserr_id); len -= sizeof(genom_syserr_id);
    memcpy(&exdetail, p, sizeof(exdetail));
    len -= sizeof(exdetail);
    genom_syserr(&exdetail);
    errnoSet(ERROR);
    size = ERROR;
    s = 0;
  } else if (!strcmp(p, genom_unkex_id)) {
    genom_unkex_detail exdetail;
    p += sizeof(genom_unkex_id); len -= sizeof(genom_unkex_id);
    memcpy(&exdetail, p, sizeof(exdetail));
    len -= sizeof(exdetail);
    genom_unkex(&exdetail);
    errnoSet(ERROR);
    size = ERROR;
    s = 0;
  } else {
    genom_unkex_detail d;
    strncpy(d.what, p, sizeof(d.what)); d.what[sizeof(d.what)-1] = *"";
    genom_unkex(&d);
    errnoSet(ERROR);
    return ERROR;
  }

  if (s) {
    switch(s) {
      case ENOMSG: errnoSet(S_gcomLib_SMALL_DATA_STR);
      case ENOMEM: errnoSet(S_gcomLib_MALLOC_FAILED);
      default: errnoSet(S_gcomLib_NOT_A_LETTER);
      }
    return ERROR;
  }

  if (len != 0) {
    errnoSet(S_gcomLib_SMALL_LETTER);
    return ERROR;
  }
  return size;
}

void
genom_<"$comp">_client_<"[$s name]">_init_input(
  struct genom_<"$comp">_<"[$s name]">_input *in)
{
<'    foreach p [$s parameters in inout] {'>
  genom_tinit_<"[[$p type] mangle]">(
    <"[[$p type] pass reference in->[$p name]]">);
<'    }'>

  /* set default value */
  genom_<"$comp">_client_<"[$s name]">_json_scan(in,
                  "<"[string map {\" \\\"} [json-param-defval $s in inout]]">",
                  NULL);
}

void
genom_<"$comp">_client_<"[$s name]">_init_output(
  struct genom_<"$comp">_<"[$s name]">_output *out)
{
<'    foreach p [$s parameters inout out] {'>
  genom_tinit_<"[[$p type] mangle]">(
    <"[[$p type] pass reference out->[$p name]]">);
<'    }'>
}

void
genom_<"$comp">_client_<"[$s name]">_fini_input(
  struct genom_<"$comp">_<"[$s name]">_input *in)
{
<'    foreach p [$s parameters in inout] {'>
  genom_tfini_<"[[$p type] mangle]">(
    <"[[$p type] pass reference in->[$p name]]">);
<'    }'>
}

void
genom_<"$comp">_client_<"[$s name]">_fini_output(
  struct genom_<"$comp">_<"[$s name]">_output *out)
{
<'    foreach p [$s parameters inout out] {'>
  genom_tfini_<"[[$p type] mangle]">(
    <"[[$p type] pass reference out->[$p name]]">);
<'    }'>
}

<'}'>
