<'
# Copyright (c) 2011-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Feb 29 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>

#include "commonStructLib.h"
#include "h2evnLib.h"
#include "taskLib.h"

#include "<"$comp">_parameters.h"
#include "<"$comp">_msglib.h"
#include "<"$comp">_activity.h"
#include "<"$comp">_control_task.h"
#include "typecopy.h"


/* --- local data ---------------------------------------------------------- */

#define CNTRL_TASK_MIN_STACK_SIZE 4096
#define EXEC_TASK_MIN_STACK_SIZE  4096

/* cslib server with ids pointer, for reentrant csLib callbacks */
struct CS_SERV_IDS {
  CS_SERV csserv;
  struct genom_<"$comp">_cntrlstr *cids;
};

static struct genom_<"$comp">_cntrlstr *global_cids;

<'foreach s [$component services] {'>
static void	<"$comp">_<"[$s name]">_rqstcb(SERV_ID csserv, int sid);
<'}'>
static void *	<"$comp">_cntrl_task(void *);

struct genom_<"$comp">_cntrlstr *
genom_<"$comp">_cids(void) { return global_cids; }


/* <"[--- genom_${comp}_init --------------------------------------------]"> */

struct genom_<"$comp">_cntrlstr *
genom_<"$comp">_init(void)
{
  struct genom_<"$comp">_cntrlstr *cids;
  char tname[64];
  STATUS s;
  int i;

  /* create internal data structure */
  if (commonStructCreate(sizeof(*cids), (void **)&cids) != OK)  {
    genom_log_warn(1, "cannot create internal data structure");
    return NULL;
  }
  global_cids = cids;
  cids->shutdown = 0;
  snprintf(cids->mbox_name, sizeof(cids->mbox_name), "%s", genom_instance);
  cids->csserv = NULL;
  cids->taskid = ERROR;
  cids->sync = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
  if (!cids->sync) {
    genom_log_warn(1, "cannot create control semaphore");
    goto error;
  }

<'if {![catch {$component ids} ids]} {'>
  genom_tinit_<"[$ids mangle]">(&cids->ids);
<'}'>
<'foreach p [$component ports] {'>
  genom_tinit_<"$comp">_<"[$p name]">_port(&cids->ports.<"[$p name]">);
<'}'>
<'foreach r [$component remotes] {'>
  genom_tinit_<"$comp">_<"[$r name]">_remote(&cids->remotes.<"[$r name]">);
<'}'>
<'foreach t [$component tasks] { '>
  cids->tasks.<"[$t name]">.taskid = ERROR;
  cids->tasks.<"[$t name]">.runnable = 0;
  cids->tasks.<"[$t name]">.sync = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
  if (!cids->tasks.<"[$t name]">.sync) {
    genom_log_warn(1, "cannot create control semaphore");
    goto error;
  }
<'}'>

  cids->activities.last = 0;
  for(i=0; i<sizeof(cids->activities.i)/sizeof(cids->activities.i[0]); i++) {
    cids->activities.i[i].any.status = ACT_VOID;
  }
  memset(cids->run_map, 0, sizeof(cids->run_map));

  /* create posters */
  s = genom_metadata_<"$comp">_init(cids);
  if (s) goto error;
<'foreach p [$component ports simple] {'>
  if (genom_<"$comp">_<"[$p name]">_open()) goto error;
<'}'>
  s = genom_state_<"$comp">_init(cids);
  if (s) goto error;
  if (genom_<"$comp">_genom_metadata_write()) goto error;


  /* spawn execution tasks */
<'foreach t [$component tasks] {'>
<'  if {[catch {[$t priority] value} priority]} { set priority 255 }'>
<'  if {[catch {[$t stack] value} stack]} { set stack [expr {1<<23}] }'>
  snprintf(tname, sizeof(tname), "%s/%s", genom_instance, "<"[$t name]">");
  if (taskSpawn2(tname, <"$priority">,
                 VX_FP_TASK, EXEC_TASK_MIN_STACK_SIZE + <"$stack">,
                 genom_<"$comp">_<"[$t name]">_exec_task, cids) == ERROR) {
    genom_log_warn(1, "cannot spawn %s exec task", "<"[$t name]">");
    goto error;
  }
  s = semTake(cids->tasks.<"[$t name]">.sync, WAIT_FOREVER);
  s = commonStructTake(cids);
  if (cids->tasks.<"[$t name]">.taskid == ERROR) {
    genom_log_warn(0, "%s exec task failed", "<"[$t name]">");
    s = commonStructGive(cids);
    goto error;
  }
  s = commonStructGive(cids);
<'}'>

  /* spawn control task */
  snprintf(tname, sizeof(tname), "%s", genom_instance);
  if (taskSpawn2(tname, 10 /* priority */,
                 VX_FP_TASK, <"$COMP">_MAX_RQST_SIZE + <"$COMP">_MAX_REPLY_SIZE
                 + CNTRL_TASK_MIN_STACK_SIZE /*size*/,
                 <"$comp">_cntrl_task, cids) == ERROR) {
    genom_log_warn(1, "cannot spawn control task");
    goto error;
  }
  s = semTake(cids->sync, WAIT_FOREVER);
  s = commonStructTake(cids);
  if (cids->taskid == ERROR) {
    genom_log_warn(0, "control task failed");
    s = commonStructGive(cids);
    goto error;
  }
  s = commonStructGive(cids);

  genom_log_info("setup and running");
  return cids;

error:
  genom_<"$comp">_fini(cids);
  return NULL;
}


/* <"[--- genom_${comp}_fini --------------------------------------------]"> */

void
genom_<"$comp">_fini(struct genom_<"$comp">_cntrlstr *cids)
{
  STATUS s;

  s = commonStructTake(cids);
  if (s != OK) assert(!"cannot lock ids - giving up");
  cids->shutdown = 1;
  genom_<"$comp">_schedule_exec(cids);
  genom_<"$comp">_schedule_cntrl(cids);
  s = commonStructGive(cids);

<'foreach t [$component tasks] {'>
  if (cids->tasks.<"[$t name]">.taskid != ERROR) {
    s = semTake(cids->tasks.<"[$t name]">.sync, WAIT_FOREVER);
  }
<'}'>
  if (cids->taskid != ERROR) {
    s = semTake(cids->sync, WAIT_FOREVER);
  }

<'foreach t [$component tasks] {'>
  if (cids->tasks.<"[$t name]">.sync)
    s = semDelete(cids->tasks.<"[$t name]">.sync);
<'}'>
  if (cids->sync)
    s = semDelete(cids->sync);

<'foreach p [$component ports] {'>
  genom_<"$comp">_<"[$p name]">_delete();
<'}'>

<'foreach r [$component remotes] {'>
  genom_tfini_<"$comp">_<"[$r name]">_remote(&cids->remotes.<"[$r name]">);
<'}'>
<'foreach p [$component ports] {'>
  genom_tfini_<"$comp">_<"[$p name]">_port(&cids->ports.<"[$p name]">);
<'}'>
<'if {![catch {$component ids} ids]} {'>
  genom_tfini_<"[$ids mangle]">(&cids->ids);
<'}'>
  commonStructDelete(cids);
  genom_log_info("shutdown complete");
}


/* <"[--- genom_${comp}_schedule_cntrl ----------------------------------]"> */

void
genom_<"$comp">_schedule_cntrl(struct genom_<"$comp">_cntrlstr *cids)
{
  if (cids->taskid == ERROR) return;
  h2evnSignal(cids->taskid);
}


/* <"[--- genom_${comp}_schedule_exec -----------------------------------]"> */

void
genom_<"$comp">_schedule_exec(struct genom_<"$comp">_cntrlstr *cids)
{
<'foreach t [$component tasks] {'>
  if (cids->tasks.<"[$t name]">.taskid != ERROR) {
    if (!cids->tasks.<"[$t name]">.runnable) {
<'  if {[catch {$t period}]} {'>
      h2evnSignal(cids->tasks.<"[$t name]">.taskid);
<'  }'>
      cids->tasks.<"[$t name]">.runnable = 1;
    }
  }
<'}'>
}


/* <"[--- ${comp}_cntrl_task(void) --------------------------------------]"> */

static void *
<"$comp">_cntrl_task(void *data)
{
  struct genom_<"$comp">_cntrlstr *cids = data;
  struct genom_activity *a;
  int id;
  int e;

  commonStructTake(cids);

  /* create requests reception mailbox */
  if (csMboxInit(cids->mbox_name, <"$COMP">_MBOX_RQST_SIZE, 0) != OK) {
    genom_log_warn(1, "cannot create server mailbox");
    commonStructGive(cids);
    semGive(cids->sync);
    return NULL;
  }

  /* configure as a server */
  if (csServInitN(
	<"$COMP">_MAX_RQST_SIZE,
	<"$COMP">_MAX_REPLY_SIZE >= <"$COMP">_MAX_INTERMED_REPLY_SIZE ?
	<"$COMP">_MAX_REPLY_SIZE : <"$COMP">_MAX_INTERMED_REPLY_SIZE,
	<"$COMP">_NRQSTID, &cids->csserv) != OK) {
    genom_log_warn(1, "cannot initialize server mailbox");
    goto error;
  }
  /* fiddle with returned CS_SERV structure, to store ids for callbacks */
  cids->csserv = realloc(cids->csserv, sizeof(struct CS_SERV_IDS));
  ((struct CS_SERV_IDS *)cids->csserv)->cids = cids;

  /* install services callbacks */
<'foreach s [$component services] {'>
  if (csServFuncInstall(cids->csserv, <"$COMP">_<"[$s name]">_RQSTID,
                        (FUNCPTR)<"$comp">_<"[$s name]">_rqstcb) != OK) {
    genom_log_warn(1, "cannot serve service <"[$s name]">");
    goto error;
  }
<'}'>

  /* signal parent */
  cids->taskid = taskIdSelf();
  genom_log_info("spawned control task");
  semGive(cids->sync);

  while(!cids->shutdown) {
    commonStructGive(cids);
    if (h2evnSusp(0) != TRUE) {
      genom_log_warn(1, "someone did something nasty");
      genom_log_warn(0, "aborting");
      abort();
    }
    commonStructTake(cids);
    genom_log_debug("control task wake up");

    /* read reception mailbox status */
    e = csMboxStatus(RCV_MBOX);
    if (e & RCV_MBOX)  {
      if (csServRqstExec(cids->csserv) != OK) {
        genom_log_warn(1, "someone did something nasty");
        genom_log_warn(0, "aborting");
        abort();
      }
    }

    /* process pending activities */
    for(id=0; id<<"$COMP">_MAX_ACTIVITIES; id++) {
      a = &cids->activities.i[id].any;
      switch(a->status) {
        case ACT_VOID:
        case ACT_RUNNING:
        case ACT_STOPPING:
        case ACT_STOP:
          break;
        case ACT_START:
          if (!genom_<"$comp">_interrupt_reqd(cids, a)) {
            a->status = ACT_RUNNING;
            a->codel = <"$comp">_start;
            genom_<"$comp">_schedule_exec(cids);
          }
          break;
        case ACT_STOPPED:
        case ACT_ETHER:
          genom_<"$comp">_activity_report(cids, a);
          /* wake up sleeping activities */
          genom_<"$comp">_schedule_exec(cids);
          break;
      }
    }
  }

  genom_log_info("shutting down control task");

error:
  if (cids->csserv) csServEnd(cids->csserv);
  csMboxEnd();
  commonStructGive(cids);
  semGive(cids->sync);
  return NULL;
}


/* --- Service callbacks --------------------------------------------------- */

<'foreach s [$component services] {'>

static void
<"$comp">_<"[$s name]">_rqstcb(SERV_ID csserv, int rid)
{
  struct genom_<"$comp">_cntrlstr *cids = ((struct CS_SERV_IDS *)csserv)->cids;
  struct genom_<"$comp">_<"[$s name]">_activity *a;
  STATUS s;
  int delay;
  int id;

  genom_log_debug("handling request for %s", "<"[$s name]">");

  /* get an activity id */
  id = genom_<"$comp">_activity_alloc(cids);
  if (id < 0) {
    s = csServReplySend(
      csserv, rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_too_many_activities_encodex);
    if (s == ERROR) assert(!"unexpected failure");
    return;
  }

  a = &cids->activities.i[id].s_<"[$s name]">;
  assert(a->h.status == ACT_VOID);
  a->h.sid = <"$COMP">_<"[$s name]">_RQSTID;
  a->h.rid = rid;
  a->h.aid = id;
  a->h.status = ACT_START;
  genom_tinit_<"$comp">_<"[$s name]">_activity(a);

  /* decode input */
  s = csServRqstParamsGet(csserv, rid, (char *)&a->in, 0,
                          genom_<"$comp">_<"[$s name]">_decode);
  if (s == ERROR) {
    genom_log_warn(0, "invalid input for service %s", "<"[$s name]">");
    s = csServReplySend(
      csserv, rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_serialization_encodex);
    if (s == ERROR) assert(!"unexpected failure");
    goto error;
  }

  /* control codels (validate + stateless) */
  a->h.codel = genom_<"$comp">_<"[$s name]">_controlcb(cids, a);
  if (a->h.codel) {
    a->h.exdetail = (void *)genom_thrown(NULL);
    s = csServReplySend(csserv, rid, FINAL_REPLY, ERROR, (char *)&a->h, 0,
                        genom_<"$comp">_activity_encodex);
    if (s == ERROR) assert(!"unexpected failure");
    goto error;
  }

  /* interrupt incompatible activities */
  delay = genom_<"$comp">_interrupt_reqd(cids, (struct genom_activity *)a);

<'  if {[catch {$s task} t]} {'>
  /* update after/before array */
  cids->run_map[<"$COMP">_<"[$s name]">_RQSTID] = 1;

  /* send final reply */
  s = csServReplySend(csserv, rid, FINAL_REPLY, OK,
                      (char *)&a->out, 0, genom_<"$comp">_<"[$s name]">_encode);
  if (s == ERROR) {
    genom_log_warn(1, "could not send output for service %s", "<"[$s name]">");
    s = csServReplySend(
      csserv, rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_serialization_encodex);
    if (s == ERROR) {
      genom_log_warn(1, "discarding output for service %s", "<"[$s name]">");
    } else {
      genom_log_warn(0, "invalid output for service %s", "<"[$s name]">");
    }
    goto error;
  }

  if (delay) /* silence brain-dead gcc warning -Wunused-but-set */;

  genom_tfini_<"$comp">_<"[$s name]">_activity(a);
  genom_<"$comp">_activity_free(cids, a->h.aid);

  genom_<"$comp">_schedule_exec(cids); /* wake up sleeping activities */
<'  } else {'>
  /* send intermediate reply */
  s = csServReplySend(csserv, rid, INTERMED_REPLY, OK,
                      (char *)&a->h.aid, sizeof(int), NULL);
  if (s == ERROR) {
    genom_log_warn(1, "cannot acknowledge service %s", "<"[$s name]">");
  }

  if (!delay) {
    /* schedule now */
    a->h.status = ACT_RUNNING;
    a->h.codel = <"$comp">_start;
    genom_<"$comp">_schedule_exec(cids);
  }
<'  }'>

  return;

error:
  genom_tfini_<"$comp">_<"[$s name]">_activity(a);
  genom_<"$comp">_activity_free(cids, a->h.aid);
}

<'}'>
